
  /**
   * @file
   *
   * This is the file for utilities that help implement platform.
   *
   * utils.js is for users of platform.
   */

const sg0                       = require('sg-argv');
const sg                        = sg0.merge(sg0, require('sg-env'), require('@sg0/sg-smart-value'));
const _                         = require('lodash');

const ENV                       = sg.ENV();

// const useSmEvents      = !!process.env.SG_LOG_SMALL_EVENTS;
const logApiCalls       = ENV.at('RA_LOG_HOST_API');
const logApiCallsV      = ENV.at('RA_LOG_VERBOSE_HOST_API');


module.exports.decodeBodyObj    = decodeBodyObj;
module.exports.argvify          = argvify;
module.exports.argvify_smart    = argvify_smart;
module.exports.contextify       = contextify;
module.exports.contextify_smart = contextify_smart;
module.exports.contextify_Xyz   = contextify_Xyz;
module.exports.normalizeHeaders = normalizeHeaders;
module.exports.methodHasBody    = methodHasBody;
module.exports.fixResponse      = fixResponse;
module.exports.mkLogApi         = mkLogApi;
module.exports.mkLogApiV        = mkLogApiV;
module.exports.asErr            = asErr;
module.exports.noop             = noop;
module.exports.cleanLog         = cleanLog;


// ------------------------------------------------------------------------------------------------------------------------------
function decodeBodyObj(body_, event, context, {smaller}) {
  if (!_.isString(body_))    { return body_; }    /* already parsed */

  var   body = sg.safeJSONParse(body_) ||{};

  // Make much smaller sometimes
  if (smaller) {
    if (Array.isArray(body.payload) && body.payload.length > 1) {
      body = {...body, payload: [body.payload[0] ||{}, `... and ${body.payload.length} more items.`]};
    }
  }

  // // 'items' is the prefered key, but only use if not already present, but leave 'payload'
  // if (!('items' in body) && Array.isArray(body.payload)) {
  //   body.items = body.payload;
  // }

  return body;
}

// ------------------------------------------------------------------------------------------------------------------------------
function argvify(query_, body_, headers_, extras, path_, method_, stage_, event_, context ={}, extraContext ={}) {
  const event = {...(event_ ||{})};

  const query     = query_    || {};
  const body      = destringify(body_     || {});
  const headers   = headers_  || {};

  const path      = path_     || event.path         || '';
  const method    = method_   || event.httpMethod   || '';

  const stage     = stage_    || event.requestContext && event.requestContext.stage   || '';

  // TODO: the headers that we know come from nginx need to be last so they are not overridden.
  const argvs     = {method, path, ...headers, ...(extras ||{}), ...body, ...query};

  const argv = {
    ...argvs,
    __meta__: {
      query,
      body,
      path,
      method,
      stage,
      headers,

      event   : event_
    }
  };

  return [argv, {...context, ...extraContext}];
}

function destringify(body) {
  if (typeof body !== 'string') {
    return body;
  }

  return sg.safeJSONParse(body);
}

// ------------------------------------------------------------------------------------------------------------------------------
function contextify_Xyz(argv, context, ...rest) {
  var   args      = [...rest];
  const callback  = (_.isFunction(_.last(args)) && args.pop()) || noop;
  const event     = args.shift();

  return callback(null, ...contextify_smart(argv, context, event));
}

// ------------------------------------------------------------------------------------------------------------------------------
function argvify_smart(event, context, argv) {
  if (!sg.isnt(argv)) {
    if (argv && argv.__meta__ && argv.__meta__.event) {
      return [argv, context];
    }

    // We have argv, but it isnt right, fix it
    return argvify(/*query=*/null, /*body=*/null, /*headers=*/null, /*extras=*/null, /*path=*/null, /*method=*/null, /*stage=*/null, sg.or(event, argv), sg.orObj(context));
  }

  return argvify(/*query=*/null, /*body=*/null, /*headers=*/null, /*extras=*/null, /*path=*/null, /*method=*/null, /*stage=*/null, sg.orObj(event), sg.orObj(context));
}

// ------------------------------------------------------------------------------------------------------------------------------
function contextify_smart(a, context, event) {
  if (!sg.isnt(a)) {

    // Ideally, argvify has already been done
    if (a && a.__meta__ && a.__meta__.event) {
      return contextify_(a, context, event);
    }

    // Maybe a is {argv}
    if (a.argv) {
      let [argv] = argvify_smart(null, context, a.argv);
      return contextify_(argv, context, event);
    }

    // Maybe a is {event}
    if (a.event || event) {
      let [argv] = argvify_smart(a.event || event, context);
      return contextify_(argv, context, a.event || event);
    }
  }

  let [argv] = argvify_smart(a, context);
  return contextify_(argv, context, event);
}

// ------------------------------------------------------------------------------------------------------------------------------
function contextify_(argv, context, ...rest) {
  return [argv ||{}, contextify(argv, context, ...rest)];
}

// ------------------------------------------------------------------------------------------------------------------------------
function contextify(argv, context, event) {
  return { ...context,
    event   : context.event || (argv && argv.__meta__ && argv.__meta__.event) || event || {},
    argv    : context.argv  || argv  || {}
  };
}

// ------------------------------------------------------------------------------------------------------------------------------
function normalizeHeaders(headers) {
  return sg.reduceObj(headers, {}, function(m, v, k) {
    return [sg.smartKey(k), v];
  });
}

// ------------------------------------------------------------------------------------------------------------------------------
const bodies = ':put:post:';
function methodHasBody(method) {
  return bodies.indexOf(':'+ method.toLowerCase() +':') !== -1;
}

// ------------------------------------------------------------------------------------------------------------------------------
function fixResponse(resp_) {
  if (sg.isnt(resp_))   { return resp_; }

  var   resp = {...resp_};

  if (sg.modes().prod) {
    resp = _.omit(resp_, 'debug', 'dbg');
  }

  return resp;
}

// ------------------------------------------------------------------------------------------------------------------------------
function mkLogApi(modType, modName) {
  return function(msg, obj, ...rest) {
    if (!logApiCalls) { return; }

    sg.log(`LOGAPI ${modName}(RA_${modType}): ${msg}`, obj, ...rest);
  };
}

// ------------------------------------------------------------------------------------------------------------------------------
function mkLogApiV(modType, modName) {
  return function(msg, obj, ...rest) {
    if (!logApiCallsV) { return; }

    sg.log(`LOGAPI ${modName}(RA_${modType}): ${msg}`, obj, ...rest);
  };
}

// ------------------------------------------------------------------------------------------------------------------------------
function asErr(obj) {
  const squashed = sg.merge(obj);
  if (sg.firstKey(squashed)) {
    return squashed;
  }

  return null;
}

// ------------------------------------------------------------------------------------------------------------------------------
function cleanLog({argv, event, context, ...rest}) {
  var argv2     = {}, event2    = {}, context2    = {};
  var replArgv  = {}, replEvent = {}, replContext = {};     /* replicas */

  if (!sg.isnt(event)) {
    event2 = {...event};

    // event.body
    if (event.body) {
      event2 = {...event2, body: shortenJsonPayload(event2.body)};
    }
  }

  if (!sg.isnt(argv)) {
    argv2 = {...argv};

    // argv.payload
    if (argv.payload) {
      argv2 = {...argv2, payload: shortenJsonPayload(argv2.payload)};
    }

    // argv.items
    if (argv.items) {
      argv2 = {...argv2, items: shortenJsonPayload(argv2.items)};
    }
  }

  return {...rest, argv: argv2, event: event2, context: context2};
}

function shortenJsonPayload(json) {
  // If its a string, just truncate it
  if (typeof json === 'string') {
    if (json.length > 128) {
      return json.substring(0, 128);
    }

    return json;
  }

  if (Array.isArray(json)) {
    return shortenArray(json);
  }

  if (sg.isObject(json)) {
    return shortenObject(json);
  }

  return json;
}

function shortenObject(json) {
  var attempt = sg.safeJSONStringify(json);
  if (attempt.length <= 256) {
    return json;
  }

  return attempt.substr(0, 128) +'...';
}

function shortenArray(json) {
  var attempt = sg.safeJSONStringify(json);
  if (attempt.length <= 256) {
    return json;
  }

  attempt = sg.safeJSONStringify(json[0]);
  if (attempt.length < 80) {
    return [json[0], `...and ${json.length - 1} more`];
  }

  return [`${json.length} items`];
}

function crackArgs({argv, argv_, event, event_, context, context_, ...rest}) {
  var cleanArgv     = argv        || argv_;
  var cleanEvent    = event       || event_;
  var cleanContext  = context     || context_;

  return {...rest, argv: cleanArgv, event: cleanEvent, context: cleanContext};
}


// ------------------------------------------------------------------------------------------------------------------------------
function noop(){}
